"""
------------------------------------------
gpstime: utilities for converting GPS time
------------------------------------------

This module contains various functions for converting between *gps-time* (the
number of seconds since January 6, 1980) and Python's
:py:class:`datetime.datetime` objects. Since gps-time is monotonous, the
conversion should correct for all leap-seconds that occurred since the start of
gps-stime. The moment of the leap seconds is obtained from the file
leap-seconds.list, which is provided twice per year by NIST. The location of
this file can be set using the environmental variable '{}'. This file has an
expiry date, and any time conversions after this date are undefined. A warning
will be logged once if the file is expired. New versions can be found at e.g.
ftp://time.nist.gov/pub/ or http://www.iana.org/time-zones
Functions converting to and from *local time* assume the time zone '{}'.

Note that the functions in the gpstime module will be discontinued in favor of
`gwpy.time.to_gps() <https://gwpy.github.io/docs/stable/time/index.html#gwpy.time.to_gps>`_
and
`gwpy.time.from_gps() <https://gwpy.github.io/docs/stable/time/index.html#gwpy.time.from_gps>`_.
"""


# Bas Swinkels, 2013


import datetime as dt
import logging
from bisect import bisect
from os import path
import pytz
import subprocess as sp
import os

try:
    import virgotools.pvt_internal_defaults as defaults
except ImportError:
    import virgotools.pvt_external_defaults as defaults


__all__ = ['utc2gps', 'gps2utc', 'now_gps', 'lt2gps', 'gps2lt', 'gps2str', 'leap_info']


logger = logging.getLogger(__name__)

UTC = pytz.utc
LT = pytz.timezone('Europe/Rome')

# epochs, in UTC
GPS_EPOCH = dt.datetime(1980, 1, 6, 0, 0, 0, tzinfo=UTC)
_NIST_EPOCH = dt.datetime(1900, 1, 1, 0, 0, 0, tzinfo=UTC)


LEAP_KEY = 'LEAP_FILE'


# http://stackoverflow.com/a/27596917
def is_timezone_aware(t):
    return t.tzinfo is not None and t.tzinfo.utcoffset(t) is not None


def _nist2datetime(t_nist):
    """Convert string with seconds since NIST_EPOCH to datetime"""
    return _NIST_EPOCH + dt.timedelta(seconds=int(t_nist))


def read_leap_seconds_list(leap_file=''):
    """Parse leapsecond.list file, return a list with moments of leapseconds in
    datetime format, a list with the corresponding number of leapseconds,
    and the expiration date of the file."""
    if not leap_file:
        leap_file = os.environ.get(LEAP_KEY, defaults.leapfile)
        if not leap_file:
            leap_file = path.join(path.dirname(path.abspath(__file__)),
                                  "leap-seconds.list")
            logger.warn("There is no leapfile defined in the %s environmental variable "
                        "or the leapfile variable in %s/pvt_external_defaults.py. "
                        "Falling back to %s, which might be outdated" %
                        (LEAP_KEY, path.dirname(path.abspath(__file__)), leap_file))
    t_leap = []
    n_leap = []
    logger.debug('Reading leap second file %s', leap_file)
    with open(leap_file) as fid:
        for line in fid:
            if line.startswith('#$'):
                t_update = _nist2datetime(line.split()[1])

            if line.startswith('#@'):
                t_expire = _nist2datetime(line.split()[1])
            line = line.split('#', 1)[0].split()  # strip comments
            if not line:
                continue  # skip emtpy lines
            assert len(line) == 2
            t_leap.append(_nist2datetime(line[0]))
            n_leap.append(int(line[1]))

    logger.debug("Leap second database last updated on %s, expires on %s",
                 t_update.date(), t_expire.date())
    now = dt.datetime.now(UTC)
    if now > t_expire:
        logger.error("Leap second database expired on %s! please update %s",
                     t_expire.date(), leap_file)
    elif now + dt.timedelta(weeks=4) > t_expire:
        logger.warn("Leap second database will expire on %s! please update %s",
                    t_expire.date(), leap_file)
    return t_leap, n_leap, t_expire


# add some hardcoded info to module docstring
__doc__ = __doc__.format(LEAP_KEY, LT.zone)


def _lookup_i_leap(edges, value):
    """looks up index of preceding leap second"""
    i = bisect(edges, value)
    if i == 0:
        raise ValueError(
            "Leap second database is not valid before {}".format(
                leap_info.t_leap[0].date()))
    return i - 1


class LeapInfo(object):
    """class for holding leap second information, which is only calculated when
    any of its attributes are used. This should improve startup time and logs
    errors only later when actually used and handlers are already installed."""

    def __init__(self):
        self._init = False

    def __getattr__(self, name):
        # use this trick to avoid implementing everything as a property
        if self._init:
            raise AttributeError(name)
        self._run_once()
        return getattr(self, name)

    def _run_once(self):
        self._init = True
        self.t_leap, self.n_leap, self.t_expire = read_leap_seconds_list()

        n_leap_at_gps_epoch = self.n_leap[_lookup_i_leap(self.t_leap, GPS_EPOCH)]
        assert n_leap_at_gps_epoch == 19  # FIXME: move to tests
        self.n_leap_since_gps_epoch = [n - n_leap_at_gps_epoch for n in self.n_leap]

        # note that the following uses info from above, so attributes are used before
        # _run_once is finished!
        self.gps_leap = [utc2gps(t) for t in leap_info.t_leap]  # convert edges to gps

    def __repr__(self):
        return 'Known leap seconds:\n%s\nList is valid until: %s' % (
            '\n'.join(str(t.date()) for t in self.t_leap), self.t_expire.date())


leap_info = LeapInfo()


def _parse_time(*args):
    """common helper for utc2gps and lt2gps"""
    if len(args) >= 3:  # year, month, day, ...
        return dt.datetime(*args)
    elif len(args) == 1:
        arg = args[0]
        if isinstance(arg, dt.datetime):  # leave datetime unchanged
            return arg
        elif isinstance(arg, dt.date):  # default to midnight
            return dt.datetime.combine(arg, dt.time())
    raise ValueError("Cannot convert arguments to a datetime")


def utc2gps(*args):
    """
    utc2gps(utctime or year, month, day, hour, minute, second, ...)
    Convert time in UTC to GPS-time.

    Accepts any valid input to the :py:class:`datetime.datetime` constructor, or
    a datetime object.

    Parameters
    ----------
    utctime : datetime
        datetime object indicating time in UTC timezone
    year, month, ... : float
        standard arguments for the :py:class:`datetime.datetime` constructor

    Returns
    -------
    gps : float
        number of seconds since start of gps-time

    Example
    -------
    >>> utc2gps(2015, 9, 14, 9, 50, 45)
    1126259462.0
    """
    t_utc = _parse_time(*args)
    if not is_timezone_aware(t_utc):
        t_utc = UTC.localize(t_utc)
    return ((t_utc - GPS_EPOCH).total_seconds() +
            leap_info.n_leap_since_gps_epoch[_lookup_i_leap(leap_info.t_leap, t_utc)])


def gps2utc(gps):
    """
    Convert GPS time to datetime object in UTC.

    Parameters
    ----------
    gps : float
        seconds since start of gps-time

    Returns
    -------
    datetime
        time in UTC

    Example
    -------
    >>> gps2utc(1126259462)
    datetime.datetime(2015, 9, 14, 9, 50, 45, tzinfo=<UTC>)
    """
    return (GPS_EPOCH + dt.timedelta(seconds=gps -
            leap_info.n_leap_since_gps_epoch[_lookup_i_leap(leap_info.gps_leap, gps)]))


def now_gps():  # FIXME: round to int, since should not be used for timing anyway
    """
    Return current time as gps-time. Note that this function should not be used
    to accurately calculate elapsed time, this is better done with
    :func:`time.time`. In future versions, the output might get rounded to
    integer.

    Returns
    -------
    float
        number of seconds since start of GPS-time
    """
    return utc2gps(dt.datetime.now(UTC))


GPS_COMMAND = '$TIMROOT/$UNAME/TiMGpsTime.exe'


def now_gps_cli():
    """Return current GPS time as integer using command line gps tool"""
    result = sp.check_output(GPS_COMMAND, shell=True)
    return int(result.split()[-1])


def lt2gps(*args):
    """
    lt2gps(localtime or year, month, day, hour, minute, second, ...)
    Convert time in local time to GPS-time.

    Accepts any valid input to the :py:class:`datetime.datetime` constructor, or
    a datetime object.

    Parameters
    ----------
    localtime : datetime
        datetime object indicating time in local time zone
    year, month, ... : float
        standard arguments for the :py:class:`datetime.datetime` constructor

    Returns
    -------
    gps : float
        number of seconds since start of gps-time
    """
    lt = _parse_time(*args)
    if not is_timezone_aware(lt):
        lt = LT.localize(lt)
    return utc2gps(lt)  # no need to convert time-zone aware time to utc


def gps2lt(gps):
    """
    Convert GPS time to datetime object in local time.

    Parameters
    ----------
    gps : float
        seconds since start of gps-time

    Returns
    -------
    datetime
        time in local time

    Example
    -------
    >>> gps2lt(1126259462)
    datetime.datetime(2015, 9, 14, 11, 50, 45, tzinfo=<DstTzInfo 'Europe/Rome' CEST+2:00:00 DST>)
    """
    return gps2utc(gps).astimezone(LT)


def gps2str(gps):
    """
    Convert GPS time to nice string with both GPS and UTC time

    Returns
    -------
    str
        String with gps-time and corresponding UTC time in friendly format

    Examples
    --------
    >>> gps2str(1126259462)
    'gps = 1126259462 (2015-09-14 09:50:45 UTC)'
    """
    return "gps = {0} ({1})".format(
        gps, gps2utc(gps).strftime('%Y-%m-%d %H:%M:%S %Z'))


# allowable input values for start_stop_dur
DUR_MAX = 100000000  # 3 year
GPS_MIN = 700000000  # 2002
GPS_MAX = 2000000000  # 2043


def start_stop_dur(gstart, gstop_or_dur):
    """
    Guess if argument is either a gps-time or a duration.

    Calculate the start-gps, stop-gps and duration of an interval that can be
    indicated by a gps-start and either a stop-gps or a duration, with some
    assumptions about typical valid ranges. This is typically used as a helper
    function to normalize input for functions that download data or triggers.

    Parameters
    ----------
    gstart : float
        gps-time indicating start of interval
    gstop_or_dur
        either a gps-time indicating the end of an interval, or a duration in
         seconds

    Returns
    -------
    gstart : float
        gps-time indicating start of interval
    gstop : float
        gps-time indicating end of interval
    duration : float
        duration of interval in seconds
    """

    if not (GPS_MIN < gstart < GPS_MAX):
        raise ValueError("gstart probably not a valid gps time")
    if 0 <= gstop_or_dur < DUR_MAX:  # probably a duration
        dur = gstop_or_dur
        gstop = gstart + dur
    elif GPS_MIN < gstop_or_dur < GPS_MAX:  # probably a gps
        gstop = gstop_or_dur
        dur = gstop - gstart
    else:
        raise ValueError("second argument is not a valid gps time or duration")
    if gstart > gstop:
        raise ValueError("gstop is smaller than gstart")
    return gstart, gstop, dur
